#include <stdio.h>
#include <stdlib.h>

#ifdef USE_PVM
#include <pvm3.h>
#endif /* USE_PVM */

#include "petrol.h"
#include "utils.h"

static void        PopulationInit(Settings *s, Population *p);
static void        PopulationIndividualsInit(Individual *i, int size, int len);
static void        PopulationDelete(Population *p);
static void        PopulationGeneration(Settings *s, Population *p);
static void        PopulationFitness(Settings *s, Population *p, IPTR pop);
static void        PopulationInitFitness(Settings *s, Population *p, IPTR pop);

Individual *Roulette(Population *p);
void        Crossover(Population *p, Individual *p1, Individual *p2,
			     Individual *c1, Individual *c2);
int         Mutation(Population *p, int x);

void ChooseBest(Settings *s, Population *p);

void PopulationRun(Settings *s, Population *p) 
{
  int gen;

  /* Initialize the population */
  PopulationInit(s, p);

  PopulationStats(s, p, p->pop);

  /* Report on the initial population */
  ReportInitial(s, p);

  for (gen = 1; gen <= p->numGenerations; gen++) {
    /* Create a new population out of the old */
    PopulationGeneration(s, p);

    PopulationStats(s, p, p->newpop);
    
    /* Report on the new generation */
    ReportGeneration(s, p, gen);

    swap(&p->pop, &p->newpop);
  }

#ifdef USE_PVM
  /* Cleanup the slave processes */
  PVMFitnessExit();
#endif

  /* Report final details */
  ReportDone(s, p);

}

void PopulationInit(Settings *s, Population *p) {
  p->pop = (Individual *)malloc(sizeof(Individual) * 
				p->popSize * p->lambda);
  if (!p->pop) {
    syserror("PopulationInit");
  }
  p->newpop = (Individual *)malloc(sizeof(Individual) * 
				   p->popSize * p->lambda);
  if (!p->newpop) {
    syserror("PopulationInit");
  }

  p->numCrossovers = 0;
  p->numMutations  = 0;

  PopulationIndividualsInit(p->pop, p->popSize * p->lambda, 
			    p->chromosomeLen);
  PopulationIndividualsInit(p->newpop, p->popSize * p->lambda, 
			    p->chromosomeLen);

  PopulationInitFitness(s, p, p->pop);
}

void PopulationIndividualsInit(Individual *i, int size, int len) {
  int n, m;

  for (n = 0; n < size; n++) {
    (i + n)->chrom = (ChromType *)malloc(sizeof(ChromType) * len);
    if (!(i + n)->chrom) {
      syserror("PopulationIndividualsInit");
    }
    (i + n)->chromLen = len;
    (i + n)->fitness = 0.0;

    for (m = 0; m < i->chromLen; m++) {
      (i + n)->chrom[m] = flip(0.5);
    }
    AppInitChrom((i + n)->chrom, i->chromLen);

  }

}

void PopulationDelete(Population *p) {
  int         n;
  Individual *pop;

  if (p == NULL)
    return;

  pop = p->pop;
  for (n = 0; n < p->popSize; n++) {
    free((pop + n)->chrom);
  }
  free(pop);
  p->pop = NULL;

}
void PopulationGeneration(Settings *s, Population *p) {
  Individual *p1, *p2;
  int i;
  int lambda;

  if (!p->newpop) {
    syserror("PopulationGeneration");
  }

  /* The famous assumption of an even population takes place here, too */
  lambda = p->lambda;
  if(lambda == 1)
    {
      for (i = 0; i < p->popSize; i += 2) {
	p1 = Roulette(p);
	p2 = Roulette(p);
	Crossover(p, p1, p2, (p->newpop + i), (p->newpop + i + 1));
      }
      PopulationFitness(s, p, p->newpop );
    } 
  else if (lambda > 1) 
    {
      for (i = p->popSize; i < p->popSize * lambda; i += 2) {
	p1 = Roulette(p);
	p2 = Roulette(p);
	Crossover(p, p1, p2, (p->pop + i), (p->pop + i + 1));
      }
      PopulationFitness(s, p, p->pop);
      ChooseBest(s, p);
    } 
  else 
    {
      syserror("PopulationGeneration: bad lambda value"); 
    }
}

void PopulationFitness(Settings *s, Population *p, IPTR pop)
{

#ifdef USE_PVM
  fprintf(stdout, "lambda = %d\n", p->lambda);
  if(p->lambda == 1) 
    PVMFitness(s, p, pop);
  else
    PVMLambdaFitness(s, p, p->pop);

#else 

  if(p->lambda == 1) {
    /* pop = p->newpop; */
    for (n = 0; n < p->popSize; n++) {
      (pop + n)->fitness = Fitness(pop + n);
    }
  }else if (lambda > 1){
    /* pop = p->pop; */
    for(n = p->popsize; n < p->popsize * p->lambda; n++) {
      (pop + n)->fitness = Fitness(pop + n);
    } 

  }
#endif /* USE_PVM */


}

void PopulationInitFitness(Settings *s, Population *p, IPTR pop)
{

#ifdef USE_PVM

    PVMFitness(s, p, pop);

#else 

    /* pop = p->newpop; */
    for (n = 0; n < p->popSize; n++) {
      (pop + n)->fitness = Fitness(pop + n);
    }
    
#endif /* USE_PVM */
    

}

